#include "geom_ext.h"
#include "geom.h"

Real v3_addcomp(Vector3 v)
{
  return  (v.x + v.y + v.z);
}

Vector3 v3_interp(Vector3 u, Vector3 v, int t)
{
  Vector3 a, b, c;
  a = v3_scale(1-t,u);
  b = v3_scale(t,v);
  c = v3_add(a,b);
  return c;
}

Vector3 v3_projnorm(Vector3 u, Vector3 v)
{
  int p;
  Vector3 a,b;
  p = v3_dot(u,v);
  a = v3_scale(p,v);
  b = v3_sub(a,b);
  return b;
}


int Matrix_Inv(float *r,float *m,int n)
{
	 int i,j,k,l;
	 float max,tmp,t;

	 /* identitee dans r */
	 for(i=0;i<n*n;i++) r[i]=0;
	 for(i=0;i<n;i++) r[i*n+i]=1;
	 
	 for(j=0;j<n;j++) {
			
			/* recherche du nombre de plus grand module sur la colonne j */
			max=m[j*n+j];
			k=j;
			for(i=j+1;i<n;i++)
				if (fabs(m[i*n+j])>fabs(max)) {
					 k=i;
					 max=m[i*n+j];
				}

      /* non intersible matrix */
      if (max==0) return 1;

			
			/* permutation des lignes j et k */
			if (k!=j) {
				 for(i=0;i<n;i++) {
						tmp=m[j*n+i];
						m[j*n+i]=m[k*n+i];
						m[k*n+i]=tmp;
						
						tmp=r[j*n+i];
						r[j*n+i]=r[k*n+i];
						r[k*n+i]=tmp;
				 }
			}
			
			/* multiplication de la ligne j par 1/max */
			max=1/max;
			for(i=0;i<n;i++) {
				 m[j*n+i]*=max;
				 r[j*n+i]*=max;
			}
			
			
			for(l=0;l<n;l++) if (l!=j) {
				 t=m[l*n+j];
				 for(i=0;i<n;i++) {
						m[l*n+i]-=m[j*n+i]*t;
						r[l*n+i]-=r[j*n+i]*t;
				 }
			}
	 }

	 return 0;
}


/* inversion of a 4x4 matrix */


void m4_inverse(Matrix4 *a,Matrix4 *b) /*calcula a inversa de b e poe em a*/
{
  Matrix4 tmp;
  float *r;
  memcpy(&tmp, b, 16*sizeof(float));
  /*tmp=*b;*/
  r[0] = a->r1.x;   r[1] = a->r2.x;   r[2] = a->r3.x;   r[3] = a->r4.x;
  r[4] = a->r1.y;   r[5] = a->r2.y;   r[6] = a->r3.y;   r[7] = a->r4.y;
  r[8] = a->r1.z;   r[9] = a->r2.z;   r[10] = a->r3.z;  r[11] = a->r4.z;
  r[12] = a->r1.w;  r[13] = a->r2.w;  r[14] = a->r3.w;  r[15] = a->r4.w;


  Matrix_Inv(r,&tmp,4);

}




