/*
* dv_win32.c
* device primitives for Windows 3.1/ Win32
* Carlos Henrique Levy & Luiz Henrique de Figueiredo
* 29 Sep 92
* Modificado : Paulo Mattos
* 20 Mar 98
*/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <windows.h>
#include "gp.h"

/* window's initial size and position */

static int NX  = 680;
static int NY  = 500;

#define X0      0
#define Y0      0
#define XMAX    (NX - 1)
#define YMAX    (NY - 1)
 
/* defines rgb color type */
typedef struct  
{
  int r, g, b;
} color_rgb;

/* defines basic color type */
typedef struct 
{
  char      name[20];  /* color name  */
  color_rgb rgb;       /* color value */
} basic_color_type;

/* create array of basic colors */
static basic_color_type basic_color[] = 
{
  { "black",     {0   , 0      , 0  } },
  { "blue",      {0   , 0      , 255} },
  { "brown",     {165 , 42     , 42 } },
  { "cyan",      {0   , 255    , 255} },
  { "grey",      {192 , 192    , 192} },
  { "green",     {0   , 255    , 0  } },
  { "lightgrey", {211 , 211    , 211} },
  { "magenta",   {255 , 0      , 255} },
  { "red",       {255 , 0      , 0  } },
  { "white",     {255 , 255    , 255} },
  { "yellow",    {255 , 255    , 0  } }
};

/* calculate number of basic colors */
#define NUM_BASIC_COLORS  (int)(sizeof(basic_color)/sizeof(basic_color_type))

/* total number of colors supported at a given time */
#define NUM_COLORS  256

/* create color map */
static struct 
{
  color_rgb rgb;          /* rgb value for color    */
  HBRUSH    brush;        /* brush  for rgb color   */
  HPEN      pen;          /* pen for rgb color      */
} color_map[NUM_COLORS];
       
/* hold DV var's */
static struct 
{
  HANDLE hinst;           /* aplication instance        */
  HWND   hwnd;            /* aplication window's handle */
  LPSTR  classname;       /* class name for window      */
  POINT  cursorpos;       /* current mouse position     */
  int    color;           /* current color              */
  char   font[256];       /* current font               */
  int    event;           /* event ?                    */
  char   report[256];     /* event string               */
  int    marksize;        /* mark size                  */
  char   marktype[20];    /* mark type                  */
  struct 
  {
    int num_pt;           /* number of points in poly   */
    int type;             /* poly type                  */ 
    int Fx,Fy;            /* poly first point           */
    int Lx,Ly;            /* poly last point            */
  } poly;
} DV = 
{
  (HANDLE)0,
  (HWND)0,
  "gp_window_class",
  {0,0},
  -1,
  "",
  0, 
  "",
  1,
  0,
  ".",
  {0,0}
};  

/* hold canvas info */
static struct 
{
  HDC     hdc;      /* Handle of current device context (DC)                */
  HDC     cl_hdc;   /* Handle of Client Area device context                 */
  int     dd_on;    /* Double-Buffer is on ?                                */
  HDC     dd_hdc;   /* Handle of Double-Buffer device context , if avaiable */
  HBITMAP dd_bmp;   /* Handle of Double-Buffer bitmap , if avaiable         */
  Box     size;     /* Device context size                                  */
} Canvas = 
{ 
  0,0,0,0,0, 
  {0.0,0.0,0.0,0.0,0.0,0.0}
};

#define MAX_FMT_STRING_SIZE 512

/* -------------------------------------------------------- */
static void __cdecl DebugPrintf (const char* fmt, ... )
{
  char Buff[MAX_FMT_STRING_SIZE];

  /* puts msg in Buff */
  vsprintf(Buff, fmt, (char *)(&fmt + 1));

  /* writes to Debug Console */
  OutputDebugString(Buff);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
static void __cdecl ShowWin32Error (const char* title, ...)
{
  char Buff[MAX_FMT_STRING_SIZE];
  char Msg[MAX_FMT_STRING_SIZE];
  char Title[MAX_FMT_STRING_SIZE];

  /* formats error msg */
  FormatMessage
  ( 
    FORMAT_MESSAGE_FROM_SYSTEM,    
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* default language */
    Buff,
    MAX_FMT_STRING_SIZE - 1,
    NULL
  );
  sprintf(Msg, "WIN32 : %s", Buff);

  /* format title */
  vsprintf(Title, title, (char *)(&title + 1));

  /* opens a dialog with error msg */
  MessageBox(NULL, Msg, Title, MB_OK | MB_ICONSTOP | MB_TASKMODAL | MB_TOPMOST);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
static void set_viewport (HDC hdc, int xmax, int ymax)
{
  /* set canvas size */
  Canvas.size.xmin  = X0;
  Canvas.size.xmax  = xmax;
  Canvas.size.ymin  = ymax;
  Canvas.size.ymax  = Y0;
  Canvas.size.xu    = xmax + 1;
  Canvas.size.yu    = ymax + 1;
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
static void clip (HDC hdc, int on)
{
  static HRGN clip_reg = 0;
  
  if (!hdc)
    return;

  if (!on)
  {
    clip_reg = CreateRectRgn (0, 0, 0, 0);
    GetClipRgn(hdc, clip_reg); 
    SelectClipRgn(hdc, NULL);
  }
  else if (clip_reg != 0)
  {
    SelectClipRgn(hdc, clip_reg);
    DeleteObject(clip_reg);
    clip_reg = 0;
  }
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
LONG FAR PASCAL windowproc (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
  int   i;
  char *r;
  
  switch (msg)
  {  
    case WM_CREATE: /* Creates window */

      #ifdef _DEBUG
      DebugPrintf("<dv_win32.c> : windowproc -> WM_CREATE \n");
      #endif

      {
        RECT rect; 
        Canvas.hdc = Canvas.cl_hdc = GetDC (hwnd);
        GetClientRect (hwnd, (LPRECT) &rect); /* get canvas size */
        set_viewport(Canvas.hdc, rect.right, rect.bottom);

        #ifdef _DEBUG
        DebugPrintf("<dv_win32.c> : WM_CREATE -> Client %d x %d \n", rect.right, rect.bottom);
        GetWindowRect (hwnd, (LPRECT) &rect); /* get canvas size */
        DebugPrintf("<dv_win32.c> : WM_CREATE -> Window %d x %d \n", rect.right - rect.left + 1, rect.bottom - rect.top + 1);
        #endif
      }
      for (i = 0; i < NUM_COLORS; i++) /* Init color map */
      {
        color_map[i].brush = (HBRUSH)0;
        color_map[i].pen   = (HPEN)0;
      }
      break;
          
    case WM_DESTROY: /* Destroys window */
      
      #ifdef _DEBUG
      DebugPrintf("<dv_win32.c> : windowproc -> WM_DESTROY \n");
      #endif
      
      /* deselect current brush and pen */
      SelectObject(Canvas.hdc, GetStockObject(WHITE_BRUSH));
      SelectObject(Canvas.hdc, GetStockObject(WHITE_PEN));
      SetClassLong(DV.hwnd, GCL_HBRBACKGROUND, (LONG)GetStockObject(WHITE_BRUSH));
        
      for (i = 0; i < NUM_COLORS;i++) /* delete all brushes and pens */
      {
        if (color_map[i].brush) /* delete brush */
        {
          DeleteObject(color_map[i].brush);
          color_map[i].brush = (HBRUSH)0;
        }
        if (color_map[i].pen) /* delete pen   */
        {
          DeleteObject(color_map[i].pen);
          color_map[i].pen = (HPEN)0;
        }
      }
      
      DV.hwnd       = (HWND)0;
      Canvas.cl_hdc = (HDC)0;
      PostQuitMessage(0); /* send WM_QUIT */
      
      break;
   /*     
    case WM_SIZING:

      #ifdef _DEBUG
      DebugPrintf("<dv_win32.c> : WM_SIZING \n");
      #endif
      break;
    */

    case WM_SIZE:

      if (wp != SIZE_MINIMIZED)
      {
        #ifdef _DEBUG
        DebugPrintf("<dv_win32.c> : WM_SIZE -> %d x %d \n", LOWORD(lp), HIWORD(lp));
        #endif
        DV.event = 1;
        strcpy(DV.report, "s+");
        DV.cursorpos.x = LOWORD(lp);
        DV.cursorpos.y = HIWORD(lp);
        if (Canvas.dd_on) 
        {
          set_viewport(Canvas.cl_hdc, LOWORD(lp), HIWORD(lp));
          dvdoublebuffer(0);  
          dvdoublebuffer(1);
        }
        set_viewport(Canvas.hdc, LOWORD(lp), HIWORD(lp));
      }
      break;
    
    case WM_ERASEBKGND:
      {
        int ret;
        clip((HDC)wp, 0); /* turns clipping off */
        ret = DefWindowProc(hwnd, msg, wp, lp);
        clip((HDC)wp, 1); /* turns clipping on */

        return ret;
      }
      break;
    
    case WM_PAINT:

      if (DV.event) break;
 
      {
        DV.event = 1;
        strcpy(DV.report, "r+");
        ValidateRect(DV.hwnd, NULL);
        #ifdef _DEBUG
        DebugPrintf("<dv_win32.c> : WM_PAINT \n");
        #endif
      }
      break;
      
    case WM_MOUSEMOVE:
     
      r = DV.report;
      *r++ = 'm';
      if (wp & MK_LBUTTON) *r++= '1';
      if (wp & MK_MBUTTON) *r++= '2';
      if (wp & MK_RBUTTON) *r++= '3';
      *r++= '+';
      if (wp & MK_SHIFT)   *r++= 'S';
      if (wp & MK_CONTROL) *r++= 'C';
      *r='\0';
      DV.cursorpos.x = LOWORD(lp);
      DV.cursorpos.y = HIWORD(lp);
      DV.event = 1;
      #ifdef _DEBUG
      //DebugPrintf("<dv_win32.c> : WM_MOUSE -> x : %d y : %d \n", DV.cursorpos.x, DV.cursorpos.y);
      #endif
      break;

    case WM_LBUTTONDOWN:
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_MBUTTONUP:
    case WM_RBUTTONUP:
      r = DV.report;
      *r++ = 'b';
      if      (msg == WM_LBUTTONDOWN) { *r++= '1'; *r++= '+';}
      else if (msg == WM_MBUTTONDOWN) { *r++= '2'; *r++= '+';}
      else if (msg == WM_RBUTTONDOWN) { *r++= '3'; *r++= '+';}
      else if (msg == WM_LBUTTONUP)   { *r++= '1'; *r++= '-';}
      else if (msg == WM_MBUTTONUP)   { *r++= '2'; *r++= '-';}
      else if (msg == WM_RBUTTONUP)   { *r++= '3'; *r++= '-';}
      if (wp & MK_SHIFT)   *r++= 'S';
      if (wp & MK_CONTROL) *r++= 'C';
      *r='\0';
      DV.cursorpos.x = LOWORD(lp);
      DV.cursorpos.y = HIWORD(lp);
      DV.event = 1;
      break;

    case WM_CHAR:              

      r = DV.report;
      *r++ = 'k';
      *r++ =  wp;
      *r++ = '+';
      if (GetKeyState (VK_SHIFT) < 0)    *r++= 'S';
      if (GetKeyState (VK_CONTROL) < 0)  *r++= 'C';
      *r='\0';
      DV.cursorpos.x = LOWORD(GetMessagePos());
      DV.cursorpos.y = HIWORD(GetMessagePos());
      ScreenToClient(hwnd, &DV.cursorpos);
      DV.event = 1;
      break;

    default:
      return DefWindowProc(hwnd,msg,wp,lp);
 }
 
 return 0;
}
/* -------------------------------------------------------- */
            
/* -------------------------------------------------------- */
int PASCAL WinMain (HANDLE hinstance, HANDLE hprev, LPSTR cl, int show)
{
  #define  max_args 50
  int      argc,i;
  char     *argv[max_args],*cl_full,*pa;
  WNDCLASS wc;
  
  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : WinMain -> Start \n");
  #endif

  DV.hinst = hinstance; /* saves app instance */
             
  /* creates window class */
  if (!hprev)
  {
    wc.lpszClassName = DV.classname;                       /* window class     */
    wc.hInstance     = hinstance;                          /* instance         */
    wc.lpfnWndProc   = (LPVOID) windowproc;                /* window function  */
    wc.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC; /* window style     */
    wc.lpszMenuName  = (LPSTR) NULL;                       /* no menu          */
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);        /* mouse cursor     */
    wc.hIcon         = LoadIcon (NULL, IDI_APPLICATION);   /* default icon     */
    wc.hbrBackground = GetStockObject(WHITE_BRUSH);        /* white background */
    wc.cbClsExtra    = 0;                                  /* no extra bytes   */
    wc.cbWndExtra    = 0;                                  /* no extra bytes   */

    if (!RegisterClass (&wc))
    {                                    
      ShowWin32Error ("GP Error <WinMain>");
      return 0;
    }
  }
     
  /* calculate "argc" and "argv" */ 
  argc    = 1; 
  argv[0] = cl_full = GetCommandLine(); /* Gets the full command line */
  if ( (strlen(cl) > 0) && (pa = strstr(cl_full, cl)) != NULL )
  {
    cl_full = pa - 1; /* space */
    for (;;argc++)
    {
      if ((pa = strchr(cl_full, ' ')) != NULL)
      {
        assert(argc < max_args);
        *pa = '\0';
        argv[argc] = cl_full = pa + 1;
      }
      else
        break;
    }
  }
  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : WinMain -> argc = %d \n", argc);
  for (i = 0;i < argc;i++)
    DebugPrintf("<dv_win32.c> : WinMain -> argv[%d] = %s \n", i, argv[i]);
  #endif
  
  /* calls aplication "main" function with "argc" and "argv" */ 
  gpmain(argc, argv);
   
  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : WinMain -> End \n");
  #endif

  return 0;
}
/* -------------------------------------------------------- */
  
/* -------------------------------------------------------- */
Box* dvopen(char* name)
{
  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : dvopen -> Start \n");
  #endif
     
  DV.hwnd = CreateWindow (
    DV.classname,           /* window class        */
    name,                   /* title               */
    WS_OVERLAPPEDWINDOW,    /* window style        */
    CW_USEDEFAULT,          /* x-position          */
    CW_USEDEFAULT,          /* y-position          */
    NX - 1,                 /* width               */
    NY - 1,                 /* height              */
    NULL,                   /* parent window       */
    NULL,                   /* menu                */
    DV.hinst,               /* instance of app.    */
    NULL);                  /* no extra parameters */
 
  if (!DV.hwnd) /* window creation failed ? */
  {
    ShowWin32Error ("GP Error <dvopen>");
    dvclose(0);
    return NULL;
  }

  UpdateWindow(DV.hwnd);
  ShowWindow(DV.hwnd, SW_SHOW);

  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : dvopen -> End \n");
  #endif

  return &Canvas.size;
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvclose(int wait)
{
  int x,y;
  
  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : dvclose -> Start \n");
  #endif
  
  if (DV.hwnd == 0) /* already destroyed ? */
    return;
       
  if (wait) 
    dvwait(-1);
  
  if (DV.hwnd == 0) /* already destroyed ? */
    return;

  DestroyWindow (DV.hwnd);

  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : dvclose -> After DestroyWindow(...) \n");
  #endif

  /* let's empty the message queue */
  while (dvevent(1, &x, &y) != NULL); 

  #ifdef _DEBUG
  DebugPrintf("<dv_win32.c> : dvclose -> End \n");
  #endif
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvclear(int wait)
{
  int c;

  if (wait) dvwait(-1);
  
  ShowCursor(TRUE);
  c = dvcolor(0);
  Rectangle(Canvas.hdc, 0, 0, (int)Canvas.size.xu - 1, (int)Canvas.size.yu - 1);
  dvcolor(c);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvflush(void)
{
}
/* -------------------------------------------------------- */
 
/* -------------------------------------------------------- */
void dvclip(int xmin, int xmax, int ymin, int ymax)
{
  /* creates cliping region in devices coordinates (not logical) */
  HRGN clip_reg = CreateRectRgn (xmin, ymin, xmax + 1, ymax + 1);
  
  if (Canvas.dd_on) 
    SelectClipRgn(Canvas.cl_hdc, clip_reg);
  SelectClipRgn(Canvas.hdc, clip_reg);
                         
  DeleteObject(clip_reg);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvcircle(int x, int y, int r)
{
  Arc(Canvas.hdc, x - r, y + r, x + r, y - r, x + r, y, x + r, y);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvellipse (int xmin, int ymin, int xmax, int ymax)
{
  int aux;
  if (xmin > xmax){aux=xmin; xmin=xmax; xmax=aux;}
  if (ymin > ymax){aux=ymin; ymin=ymax; ymax=aux;}
  Arc(Canvas.hdc, xmin, ymin, xmax, ymax, xmax, (ymax - ymin)/2, xmax, (ymax - ymin)/2);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvwait(int t)
{
  if (t >= 0)
  {
    Sleep(t);  /* waits "t" miliseconds */
  }
  else
  {
    for (;;) /* waits for user input */
    {
      int x,y; 
      char* s = dvevent(1, &x, &y);
    
      if (s == NULL || (s[0] != 'r' && s[0] != 's' && s[0] != 'm')) /* so must be input :) */
        break; 
    }
  }
}
/* -------------------------------------------------------- */
   
/* -------------------------------------------------------- */
void dvline(int x1, int y1, int x2, int y2)
{
  MoveToEx (Canvas.hdc, x1, y1, NULL);
  LineTo (Canvas.hdc, x2, y2);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvbox(int xmin, int xmax, int ymin, int ymax)
{
  int aux;
  if (xmin > xmax){aux = xmin; xmin = xmax; xmax = aux;}
  if (ymin > ymax){aux = ymin; ymin = ymax; ymax = aux;}
  
  /* Rectangle (Canvas.hdc, xmin, ymax, xmax, ymin);  */
  
  PatBlt (Canvas.hdc, xmin, ymin, 
          xmax - xmin + 1, ymax - ymin + 1, PATCOPY);
}  
/* -------------------------------------------------------- */

#if 0
/* -------------------------------------------------------- */
/* Compara funcoes : Rectangle e PatBlt                     */
/* -------------------------------------------------------- */
void dvbox(int xmin, int xmax, int ymin, int ymax)
{
  #define CNT 50000
  long t1,t2,i;
  
  int aux;
  if (xmin > xmax){aux = xmin; xmin = xmax; xmax = aux;}
  if (ymin > ymax){aux = ymin; ymin = ymax; ymax = aux;}
  
  t1 = clock();
  for (i = 0;i < CNT;i++)
    Rectangle (Canvas.hdc, xmin, ymax, xmax, ymin);
  t1 = clock() - t1;

  t2 = clock();
  for (i = 0;i < CNT;i++)
    PatBlt (Canvas.hdc, xmin, ymin, 
            xmax - xmin + 1, ymax - ymin + 1, PATCOPY);
  t2 = clock() - t2;

  DebugPrintf("<dv_win32.c> : dvbox -> Rectangle : %.1f\n", 
              (float)t1/CLOCKS_PER_SEC);
  DebugPrintf("<dv_win32.c> : dvbox -> PatBlt    : %.1f\n", 
              (float)t2/CLOCKS_PER_SEC);
}
/* -------------------------------------------------------- */
#endif

/* -------------------------------------------------------- */
void dvtri(int x1, int y1, int x2, int y2, int x3, int y3)
{
  int t[2*3];
  t[0] = x1; t[1] = y1;
  t[2] = x2; t[3] = y2;
  t[4] = x3; t[5] = y3;
  Polygon(Canvas.hdc, (LPPOINT) t, 3);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
int dvfont(char* name)
{
  SelectObject(Canvas.hdc, GetStockObject(DEFAULT_GUI_FONT));
  //SelectObject(Canvas.hdc, GetStockObject(SYSTEM_FONT));
  
  strcpy(DV.font, name);
   
  return 0;
}
/* -------------------------------------------------------- */
  
/* -------------------------------------------------------- */
void dvtext(int x, int y, char* s, char* m)
{
  #define at(a,b) (m[first]==a && m[first + 1]==b)
  WORD flag;
  int  old;

  if (m == NULL) 
  {   
    flag = TA_LEFT | TA_BOTTOM;
    old  = SetBkMode(Canvas.hdc, TRANSPARENT); 
  }
  else
  {
    int first = 0;
    
    if (m[0] == '!')
    {
      old = SetBkMode(Canvas.hdc, OPAQUE); 
      first = 1;
    }
    else
      old = SetBkMode(Canvas.hdc, TRANSPARENT); 
    
    if      (at('n', 0 )) flag = TA_CENTER | TA_TOP;
    else if (at('n','e')) flag = TA_RIGHT  | TA_TOP;
    else if (at('n','w')) flag = TA_LEFT   | TA_TOP;
    else if (at('s', 0 )) flag = TA_CENTER | TA_BOTTOM;
    else if (at('s','e')) flag = TA_RIGHT  | TA_BOTTOM;
    else if (at('s','w')) flag = TA_LEFT   | TA_BOTTOM;
    else if (at('e', 0 )) flag = TA_RIGHT  | TA_BASELINE;
    else if (at('w', 0 )) flag = TA_LEFT   | TA_BASELINE;
    else if (at('c', 0 )) flag = TA_CENTER | TA_BASELINE;
  }

  SetTextAlign (Canvas.hdc, flag);
  TextOut (Canvas.hdc, x, y, s, strlen(s));
  SetBkMode(Canvas.hdc, old); 
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvmark(int size, char* mark)
{
  if (size > 0) 
    DV.marksize = size;
  if (mark != NULL && *mark != 0) 
    strcpy(DV.marktype, mark);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvplot(int x, int y)
{
  int   size = DV.marksize;
  int   dx, dy;
  char* m;
  int   c = DV.color;
                     
  for (m = DV.marktype; *m != 0; m++)
  {
    switch (*m)
    {
      case '.':
        SetPixel(Canvas.hdc, x, y, RGB(color_map[c].rgb.r,color_map[c].rgb.g,color_map[c].rgb.b));
        break;
      case 'o':
        dvellipse (x - size, y + size, x + size, y - size);
        break;
      case 'O': /* with fill */
        Ellipse(Canvas.hdc, x - size, y + size, x + size, y - size);
        break;
      case '+':
        dvline (x - size, y, x + size, y);
        dvline (x, y - size, x, y + size);
        break;
      case '*':
        dx = size/2;
        dy = (int)(size*0.866);
        dvline (x - size, y, x+size, y);
        dvline (x + dx, y + dy, x - dx, y - dy);
        dvline (x + dx, y - dy, x - dx, y + dy);
        break;
      case 'x':
        dx = size;
        dy = size;
        dvline (x + dx, y + dy, x - dx, y - dy);
        dvline (x - dx, y - dy, x + dx, y + dy); /* simetrico */
        dvline (x + dx, y - dy, x - dx, y + dy);
        dvline (x - dx, y + dy, x + dx, y - dy); /* simetrico */
        break;
      case 'b':
        dvline(x - size, y - size, x - size, y + size);
        dvline(x - size, y + size, x + size, y + size);
        dvline(x + size, y + size, x + size, y - size);
        dvline(x + size, y - size, x - size, y - size);
        break;
      case 'B': /* with fill */ 
        dvbox (x - size,x + size,y - size,y + size);
        break;      
      case 'd':
        dvline (x - size, y, x, y + size);
        dvline (x, y + size, x + size, y);
        dvline (x + size, y, x, y - size);
        dvline (x, y - size, x - size, y);
        break;
      case 'D': /* with fill */
      { 
        POINT Pt[4];
        Pt[0].x = x - size; Pt[0].y = y;
        Pt[1].x = x;        Pt[1].y = y + size;
        Pt[2].x = x + size; Pt[2].y = y;
        Pt[3].x = x;        Pt[3].y = y - size;
        
        Polygon(Canvas.hdc, Pt, 4);
        break;
      }
    }
  }
}
/* -------------------------------------------------------- */
          
/* -------------------------------------------------------- */
char* dvevent(int wait, int* x, int* y)
{
  MSG msg;
  int msg_av = 0;
   
  DV.event = 0; /* waits for event */
  strcpy(DV.report, "");
  
  if (!wait)
  {
    if (PeekMessage(&msg, DV.hwnd, 0, 0, PM_REMOVE))
    {
      if (msg.message == WM_QUIT)
      {
        #ifdef _DEBUG
        DebugPrintf("<dv_win32.c> : dvevent -> WM_QUIT (PeekMessage) \n");
        #endif
        return NULL; /* WM_QUIT */
      }
      msg_av = 1;
    }
    else
      return NULL; /* No messages */
  }   
  
  do 
  {
    if (msg_av || GetMessage(&msg, DV.hwnd, 0, 0))
    { 
      msg_av = 0;
      
      #ifdef _DEBUG
      /*DebugPrintf("<dv_win32.c> : dvevent -> Before TranslateMessage(...) \n");*/
      #endif
      TranslateMessage (&msg);
 
      #ifdef _DEBUG
      /*DebugPrintf("<dv_win32.c> : dvevent -> Before DispatchMessage(...) \n");*/
      #endif
      DispatchMessage  (&msg);
      
      #ifdef _DEBUG
      /*DebugPrintf("<dv_win32.c> : dvevent -> report = %s \n", DV.report);*/
      #endif

      if (DV.event) 
      {
        if (DV.report[0] != 's') 
        {
          DPtoLP (Canvas.cl_hdc, &DV.cursorpos, 1);
        }
        *x = DV.cursorpos.x;
        *y = DV.cursorpos.y;
        break;
      }
    }
    else /* WM_QUIT */
    {
      #ifdef _DEBUG
      DebugPrintf("<dv_win32.c> : dvevent -> WM_QUIT (GetMessage) \n");
      #endif
      return NULL;
    }
  }
  while (wait);
 
  return DV.event ? DV.report : NULL;
}
/* -------------------------------------------------------- */

/* ******************************************** * 
 * *** Implements begin/point/end functions *** *      
 * ******************************************** */

/* -------------------------------------------------------- */
void dvbegin(int c)
{
  DV.poly.num_pt = 0;
  DV.poly.type   = c;
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
/* Note : No other primitives should be used between a 
          dvbegin - dvend call */
int dvpoint(int x, int y)
{
  if (DV.poly.type != 'm')
  {
    if (DV.poly.num_pt == 0)
    {
      MoveToEx(Canvas.hdc, x, y, NULL);
      DV.poly.Fx = x;
      DV.poly.Fy = y;
      BeginPath(Canvas.hdc);
    }  
    else
    {
      DV.poly.Lx = x;
      DV.poly.Ly = y;
      LineTo(Canvas.hdc, x, y);
      SetPixel(Canvas.hdc, x, y,
             RGB(color_map[DV.color].rgb.r, color_map[DV.color].rgb.g, color_map[DV.color].rgb.b)); 
    }
  }
  else
  {
    int dx = 3;
    int dy = 3;
    dvline (x + dx, y + dy, x - dx, y - dy);
    dvline (x - dx, y - dy, x + dx, y + dy); /* simetrico */
    dvline (x + dx, y - dy, x - dx, y + dy);
    dvline (x - dx, y + dy, x + dx, y - dy); /* simetrico */
  }

  return ++DV.poly.num_pt;
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvend(void)
{
  if (DV.poly.num_pt > 0)
  { 
    if (DV.poly.type != 'm')
      EndPath(Canvas.hdc);
    switch (DV.poly.type)
    {
      case 'l' :   /* open polygonal line   */
        StrokePath(Canvas.hdc); 
        break;
      case 'p' :   /* closed polygonal line */
        StrokePath(Canvas.hdc);         
        dvline(DV.poly.Lx, DV.poly.Ly, DV.poly.Fx, DV.poly.Fy);
        break;
      case 'f' :   /* filled polygonal line */
        StrokeAndFillPath(Canvas.hdc);
        break;
      case 'm' :   /* polymarker            */
        break;
      default  : 
        assert(0);
        break;
    }
  }  
}
/* -------------------------------------------------------- */

/* ********************************** * 
 * *** Implements Color functions *** *      
 * ********************************** */

/* -------------------------------------------------------- */
int dvcolor(int c)
{
  int old_color;
  
  if (c < 0 || c >= NUM_COLORS)
    return -NUM_COLORS;
   
  old_color = DV.color; 
  DV.color = c;
  SelectObject (Canvas.hdc, color_map[c].brush);
  SelectObject (Canvas.hdc, color_map[c].pen);
  SetTextColor (Canvas.hdc, 
                RGB(color_map[c].rgb.r,color_map[c].rgb.g,color_map[c].rgb.b));

  return old_color;
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
static int setpalette(int c, color_rgb color)
{  
  if (c == DV.color)
  {
    /* deselect current brush and pen */
    SelectObject(Canvas.hdc, GetStockObject(WHITE_BRUSH));
    SelectObject(Canvas.hdc, GetStockObject(WHITE_PEN));
  }
  
  /* Background color ? */
  if (c == 0)
  { 
    if (!SetClassLong(DV.hwnd, GCL_HBRBACKGROUND, (LONG)GetStockObject(WHITE_BRUSH)))
    {
      ShowWin32Error("<dv_win32.c> : setpalette -> SetClassLong");
      return 0;
    }
  }

  if (color_map[c].brush) /* delete brush */
    DeleteObject(color_map[c].brush);
  if (color_map[c].pen) /* delete pen   */
    DeleteObject(color_map[c].pen);
   
  /* create new brush and pen */
  color_map[c].brush = CreateSolidBrush (RGB(color.r, color.g, color.b));
  color_map[c].pen   = CreatePen (PS_SOLID, 1, RGB(color.r, color.g, color.b)); 
  color_map[c].rgb   = color; 
  assert(color_map[c].brush != 0);
  assert(color_map[c].pen != 0);

  /* Background color ? */
  if (c == 0)
  { 
    if (!SetClassLong(DV.hwnd, GCL_HBRBACKGROUND, (LONG)color_map[0].brush))
    {
      ShowWin32Error("<dv_win32.c> : setpalette -> SetClassLong");
      return 0;
    }
    SetBkColor(Canvas.hdc, RGB(color.r, color.g, color.b)); 
  }
  
  if (c == DV.color)
  {
    /* select new brush and pen */
    SelectObject(Canvas.hdc, color_map[c].brush);
    SelectObject(Canvas.hdc, color_map[c].pen);
  }

  /* imitates GP for UNIX */
  if (c == 0)
    dvclear(0);

  return 1;
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
int dvpalette(int c, char* name)
{
  int       i;
  color_rgb color;

  if (c < 0 || c >= NUM_COLORS)
    return 0;
  
  /* find basic color */
  for (i = 0;i < NUM_BASIC_COLORS;i++)
  {
    if (!strcmp(basic_color[i].name, name))
      break;     
  }
     
  if (i == NUM_BASIC_COLORS) /* didn't find color name */
  {
    if (!strncmp(name, "grey", 4)) /* greyXXX ? */
    {
      int v;
      sscanf(name, "grey%d", &v);
      if (v < 0 || v > 100)
        return 0;
      color.r = (int)(255*(v/100.0));
      color.g = (int)(255*(v/100.0));
      color.b = (int)(255*(v/100.0));
    }
    else
      return 0;
  }
  else 
    color = basic_color[i].rgb;

  return setpalette(c, color);
}
/* -------------------------------------------------------- */
       
/* -------------------------------------------------------- */
int dvrgb(int c, real r, real g, real b)
{
  color_rgb color;

  if (c < 0 || c >= NUM_COLORS)
    return 0;
  
  color.r = (int)(255*r);
  color.g = (int)(255*g);
  color.b = (int)(255*b);

  return setpalette(c, color);
}
/* -------------------------------------------------------- */

/* ******************************************* * 
 * *** Implements Double-Buffer functions ***  *      
 * ******************************************* */

/* -------------------------------------------------------- */
static void set_DC_attributes (void)
{
  dvcolor(DV.color);
  dvfont(DV.font);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvdoublebuffer (int on)
{
  if (on)
  {
    RECT screen_size;
   
    /* Get screen size */
    GetClientRect(GetDesktopWindow(), &screen_size);
    #ifdef _DEBUG
    DebugPrintf("<dv_win32.c> : dvdoublebuffer -> Screen is %d x %d \n", screen_size.right, screen_size.bottom);
    #endif
   
    /* Creates Double-Buffer */
    GetClientRect(DV.hwnd, &screen_size);
    #ifdef _DEBUG
    DebugPrintf("<dv_win32.c> : dvdoublebuffer -> Client is %d x %d \n", screen_size.right, screen_size.bottom);
    #endif
    if ((Canvas.dd_bmp = CreateCompatibleBitmap(Canvas.cl_hdc, screen_size.right, screen_size.bottom)) == NULL)
    {
      ShowWin32Error("<dv_win32.c> : dvdoublebuffer -> CreateCompatibleBitMap");
      return;
    }
    if ((Canvas.dd_hdc = CreateCompatibleDC(Canvas.cl_hdc)) == NULL)
    {
      ShowWin32Error("<dv_win32.c> : dvdoublebuffer -> CreateCompatibleDC");
      return;
    }
    if (SelectObject(Canvas.dd_hdc, Canvas.dd_bmp) == NULL)
    {
      ShowWin32Error("<dv_win32.c> : dvdoublebuffer -> SelectObject");
      return;
    } 
    Canvas.dd_on = 1;
    dvbackbuffer();
    dvclear(0);
  }
  else if (Canvas.dd_on)
  {
    dvfrontbuffer();
    DeleteObject(Canvas.dd_bmp);
    DeleteDC(Canvas.dd_hdc);
    Canvas.dd_bmp = (HBITMAP)0;
    Canvas.dd_hdc = (HDC)0;
    Canvas.dd_on  = 0;
  }
}  
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvswapbuffers  (void)
{
  if (!BitBlt(Canvas.cl_hdc, 0, 0, 
              (int)Canvas.size.xu - 1, (int)Canvas.size.yu - 1, 
              Canvas.dd_hdc, 0, 0, SRCCOPY))
  {
    ShowWin32Error("<dv_win32.c> : dvswapbuffers -> BitBlt");
    return;
  }
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvfrontbuffer (void)
{
  Canvas.hdc = Canvas.cl_hdc;
  if (Canvas.hdc)
    set_DC_attributes();
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvbackbuffer (void)
{
  Canvas.hdc = Canvas.dd_hdc; 
  set_viewport(Canvas.hdc, (int)Canvas.size.xu - 1, (int)Canvas.size.yu - 1);
  set_DC_attributes();
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- */
void dvbufferarea (int xmin, int xmax, int ymin, int ymax)
{

}
/* -------------------------------------------------------- */

void dvsetsize(int w, int h, int x, int y)
{
  NX = w; NY = h;
}
