#include "aux.h"


void destroy_image(Image *i)
{
  img_free(i);
  efree(i);
}

void empty_list(List *q)
{
  Item *i=q->head;
  while (i!=NULL) {
    Item *t=i;
    i=i->next;
    efree(t->d);
    efree(t);
  }
  q->head = q->tail = NULL;
}

void destroy_list(List *q)
{
  Item *i=q->head;
  while (i!=NULL) {
    Item *t=i;
    i=i->next;
    efree(t->d);
    efree(t);
  }
  efree(q);
}

void s4_destroy(Stack4 * s) {
  efree(s->mbot);
  efree(s->ibot);
  efree(s);
}


Vector4 s4_v4xform(Stack4 *s, Vector4 v)
{
  return v4_m4mult(v,  *s->mtop);
}

Vector3 s4_n3xform2(Stack4 *s, Vector3 nv)
{
  return v3_m3mult(nv, /*m4_transpose*/(*s->mtop));
}

void s4_matconc(Stack4 *s, Matrix4 *md, Matrix4 *im)
{
  *s->mtop = m4_m4prod(*s->mtop,*md);
  if ( im == (Matrix4 *)0)
    *s->itop = m4_m4prod(m4_ident(), *s->itop);
  else
    *s->itop = m4_m4prod(*im, *s->itop);
}

Real v3_max_dot(Vector3 d1, Vector3 d2)
{
  Real temp = v3_dot(d1,d2); 
  return (temp>0)?temp:0;
}


Vector3 v3_dir_unit(Vector3 a, Vector3 b)
{
  return v3_unit(v3_sub(b,a));
}


Vector4 v4_sub(Vector4 a, Vector4 b)
{
  a.x -= b.x; a.y -= b.y; a.z -= b.z; a.w -= b.w;
  return a;
}

Real v4_norm(Vector4 v)
{
  return sqrt((Real)(v.x * v.x + v.y * v.y + v.z * v.z + v.w * v.w));
}

Vector3 v4_dir_unit(Vector4 a, Vector4 b)
{
  if(a.w != 0 && b.w == 0)
    return v3_unit(v3_v4conv(b));
  else {
    if(a.w == 0 && b.w != 0)
      return v3_scale(-1, v3_unit(v3_v4conv(a)));
    else {
      if (a.w == 0 && b.w == 0)
	return v3_dir_unit(v3_v4conv(a), v3_v4conv(b));
      else
	return v3_dir_unit(v3_v4conv(a), v3_v4conv(b));
    }   
  }
}

Real v4_distance(Vector4 a, Vector4 b) {
  return v4_norm(v4_sub(a,b));
}

Real v3_dist(Vector3 a, Vector3 b) {
  return v3_norm(v3_sub(a,b));
}


Vector4 v4_add(Vector4 a, Vector4 b)
{
  a.x += b.x; a.y += b.y; a.z += b.z; a.w += b.w;
  return a;
}

Vector4 v4_scale(Real t, Vector4 v)
{
  v.x *= t; v.y *= t; v.z *= t; v.w *= t;
  return v;
}

Vector4 v4_unit(Vector4 u)
{
  Real  length = v4_norm(u);
  if(fabs(length) < EPS) {
    error("(g4_unit) zero norm\n");
    return u;
  }
  else
    return v4_scale(1.0/length, u);
}

Real v4_sqrnorm(Vector4 v)
{
  return (v.x * v.x + v.y * v.y + v.z * v.z + v.w * v.w);
}


