#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "gl.h"
#include "context.h"
#include "aux.h"
#include "rasterization.h"


#define SET_ERROR(X) if(current_context->error == GL_NO_ERROR) current_context->error = (X);

     
static GLContext * current_context;

GLContext * new_context(int w, int h)
{
  int i;
  GLContext * c;
  c=(GLContext *)malloc(sizeof(GLContext));

  /* inicializacao */

  zbuf_init(w, h);

  
  /**Begin**/
  c->into_begin_end = GL_FALSE;
  
  
  /**Matrix**/
  c->matrix_mode = GL_MODELVIEW;
  c->modelview_stack = s4_initstack(MODELVIEW_STACK_SIZE);
  c->projection_stack = s4_initstack(PROJECTION_STACK_SIZE);
  c->color_stack = s4_initstack(COLOR_STACK_SIZE);
  

  /**Vertex**/
  c->edge_flag = GL_TRUE;
  
  c->vertex_list = new_list();
  
  c->current_normal.x = 0.0;
  c->current_normal.y = 0.0;
  c->current_normal.z = 1.0;

  c->current_color.x = 1.0;
  c->current_color.y = 1.0;
  c->current_color.z = 1.0;
  

   /** Viewport **/
  c->w = w;
  c->h = h;
  c->ox = c->w/2;
  c->oy = c->h/2;
  c->n = 0.0;
  c->f = 1.0;


  /** Enable **/
  c->normalize_enabled = GL_FALSE;
  c->rescale_normal_enabled = GL_FALSE;
  c->smooth_enabled = GL_TRUE;
  c->depth_test_enabled = GL_FALSE;
  c->lighting_enabled = GL_FALSE; /*verificar*/
  c->color_material_enabled = GL_FALSE; /*verificar*/
  for(i=0;i<MAX_LIGHTS;i++)
    c->light_enabled[i] = GL_FALSE;
  

  /** Front Face **/
  c->front_sign = 1.0;
  
  /** Cull Face **/
  c->cull_sign = -1.0;
  
  /**Material Parameters**/
  c->material.ambient.x = 0.2;
  c->material.ambient.y = 0.2;
  c->material.ambient.z = 0.2;
  
  c->material.diffuse.x = 0.8;
  c->material.diffuse.y = 0.8;
  c->material.diffuse.z = 0.8;
    
  c->material.specular.x = 0.0;
  c->material.specular.y = 0.0;
  c->material.specular.z = 0.0;
    
  c->material.emissive.x = 0.0;
  c->material.emissive.y = 0.0;
  c->material.emissive.z = 0.0;
    
  c->material.specular_exponent = 0.0;
  c->material.ambient_index =  0.0;
  c->material.diffuse_index =  1.0;
  c->material.specular_index = 1.0;

  c->material.face_color_material =  GL_FRONT_AND_BACK;
  c->material.color_material_mode =  GL_AMBIENT_AND_DIFFUSE;
  
  /** Light Source Parameters **/
  for(i=0;i<MAX_LIGHTS;i++){
    c->l[i].ambient.x = 0.0;
    c->l[i].ambient.y = 0.0;
    c->l[i].ambient.z = 0.0;
    if(i==0){
      c->l[i].diffuse.x = 1.0;
      c->l[i].diffuse.y = 1.0;
      c->l[i].diffuse.z = 1.0;
      c->l[i].specular.x = 1.0;
      c->l[i].specular.y = 1.0;
      c->l[i].specular.z = 1.0;
    }
    else{
      c->l[i].diffuse.x = 0.0;
      c->l[i].diffuse.y = 0.0;
      c->l[i].diffuse.z = 0.0;
      c->l[i].specular.x = 0.0;
      c->l[i].specular.y = 0.0;
      c->l[i].specular.z = 0.0;
    }
    c->l[i].position.x = 0.0;
    c->l[i].position.y = 0.0;
    c->l[i].position.z = 1.0;
    c->l[i].position.w = 0.0;
    c->l[i].spot_direction.x = 0.0;
    c->l[i].spot_direction.y = 0.0;
    c->l[i].spot_direction.z = -1.0;
    c->l[i].spot_exponent = 0.0;
    c->l[i].spot_cutoff = 180.0;
    c->l[i].const_att = 1.0;
    c->l[i].linear_att = 0.0;
    c->l[i].quad_att = 0.0;
  }

  
  /** Lighting Model Parameters **/
  c->ambient_scene.x = 0.2;
  c->ambient_scene.y = 0.2;
  c->ambient_scene.z = 0.2;
  c->local_viewer = GL_FALSE;
  c->two_side = GL_FALSE;
  c->color_control = GL_SINGLE_COLOR;
  

  /**Image**/
  c->img = img_init(IMG_TRUECOL, w, h);
  c->triangle = plist_alloc(3,3);
  
  return c;
}

void delete_context(GLContext * c)
{
  /* destruicao */
  plist_free(c->triangle);
  img_free(c->img);
  destroy_list(c->vertex_list);
  s4_destroy(c->color_stack);
  s4_destroy(c->projection_stack);
  s4_destroy(c->modelview_stack);
  free(c);
}

GLContext * set_current_context(GLContext * c)
{
  return current_context = c;
}

GLContext * get_current_context()
{
  return current_context;
}

void save_context_image(GLContext *c, char *fname)
{
  img_write(c->img,fname,0);
}


void in_begin()
{
  GLContext * current_context;
  current_context = get_current_context();
  if(current_context->into_begin_end == GL_TRUE) {
    SET_ERROR(GL_INVALID_OPERATION); 
    return;
  }
}

/* Specify vertices or rectangles */
void glBegin(GLenum mode)
{
  in_begin();
  switch(mode){
  case GL_POINTS:
  case GL_LINE_STRIP:
  case GL_LINE_LOOP:
  case GL_LINES:
  case GL_POLYGON:
  case GL_TRIANGLE_STRIP:
  case GL_TRIANGLE_FAN:
  case GL_TRIANGLES:
  case GL_QUAD_STRIP:
  case GL_QUADS:
    current_context->begin_mode = mode;
    current_context->into_begin_end = GL_TRUE;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
  }
}

void glEnd (void)
{
  switch(current_context->begin_mode){
  case GL_POINTS:
    raster_points();
    break;
  case GL_LINE_STRIP:
    raster_line_strip();
    break;
  case GL_LINE_LOOP:
    raster_line_loop();
    break;
  case GL_LINES:
    raster_lines();
    break;
  case GL_POLYGON:
    raster_polygon();
    break;
  case GL_TRIANGLE_STRIP:
    raster_triangle_strip();
    break;
  case GL_TRIANGLE_FAN:
    raster_triangle_fan();
    break;
  case GL_TRIANGLES:
    raster_triangles();
    break;
  case GL_QUAD_STRIP:
    raster_quad_strip();
    break;
  case GL_QUADS:
    raster_quads();
    break;
  }
  empty_list(current_context->vertex_list);
  current_context->into_begin_end = GL_FALSE;
}

void glVertex3d (GLdouble x, GLdouble y, GLdouble z)
{
  Vertex_info * v_info;
  Item * t;

  v_info = (Vertex_info *)malloc(sizeof(Vertex_info));
  v_info->v.x = x;
  v_info->v.y = y;
  v_info->v.z = z;
  
  v_info->n.x = current_context->current_normal.x;
  v_info->n.y = current_context->current_normal.y;
  v_info->n.z = current_context->current_normal.z;

  v_info->cpri.x = current_context->current_color.x;
  v_info->cpri.y = current_context->current_color.y;
  v_info->cpri.z = current_context->current_color.z;

  v_info->csec.x = 0.0;
  v_info->csec.y = 0.0;
  v_info->csec.z = 0.0;
  
  t = new_item(v_info);
  append_item(current_context->vertex_list, t);

  
}

void glVertex3dv (const GLdouble *v)
{
  glVertex3d(v[0], v[1], v[2]);
}

void glVertex3f (GLfloat x, GLfloat y, GLfloat z)
{
  glVertex3d(x, y, z);
}

void glVertex3fv (const GLfloat *v)
{
  glVertex3d(v[0], v[1], v[2]);
}

void glVertex2f (GLfloat x, GLfloat y)
{
  glVertex3d(x, y, 0);
}

void glVertex2d (GLdouble x, GLdouble y)
{
  glVertex3d(x, y, 0);
}


/* Transform the current matrix */
void glRotated (GLdouble angle, GLdouble x, GLdouble y, GLdouble z)
{
  Vector3 s[3];
  Vector3 t[3];
  Vector3 u[3];
  Vector3 v;
  Matrix4 r,tmp;
  Stack4 * a;
  in_begin();

  angle = DTOR * angle;
  v = v3_unit(v3_make(x, y, z));
  
  t[0].x = v.x*v.x;
  t[0].y = v.x*v.y;
  t[0].z = v.x*v.z;

  t[1].x = v.y*v.x;
  t[1].y = v.y*v.y;
  t[1].z = v.y*v.z;

  t[2].x = v.z*v.x;
  t[2].y = v.z*v.y;
  t[2].z = v.z*v.z;

  s[0].x = cos(angle)*(1 - t[0].x);
  s[0].y = cos(angle)*(-t[0].y);
  s[0].z = cos(angle)*(-t[0].z);

  s[1].x = -t[1].x * cos(angle);
  s[1].y = cos(angle)*(1 - t[1].y);
  s[1].z = -t[1].z * cos(angle);

  s[2].x = -t[2].x * cos(angle);
  s[2].y = -t[2].y * cos(angle);
  s[2].z = cos(angle)*(1 - t[2].z);

  u[0].x = 0.0;
  u[0].y = -sin(angle)*v.z;
  u[0].z = sin(angle)*v.y;

  u[1].x = sin(angle)*v.z;
  u[1].y = 0.0  ;
  u[1].z = sin(angle)*(-v.x);

  u[2].x = sin(angle)*(-v.y);
  u[2].y = sin(angle)*v.x;
  u[2].z = 0.0;

  r.r1.x = t[0].x + s[0].x + u[0].x;
  r.r1.y = t[0].y + s[0].y + u[0].y;
  r.r1.z = t[0].z + s[0].z + u[0].z;
  r.r1.w = 0;

  r.r2.x = t[1].x + s[1].x + u[1].x;
  r.r2.y = t[1].y + s[1].y + u[1].y;
  r.r2.z = t[1].z + s[1].z + u[1].z;
  r.r2.w = 0;

  r.r3.x = t[2].x + s[2].x + u[2].x;
  r.r3.y = t[2].y + s[2].y + u[2].y;
  r.r3.z = t[2].z + s[2].z + u[2].z;
  r.r3.w = 0;

  r.r4.x = 0;
  r.r4.y = 0;
  r.r4.z = 0;
  r.r4.w = 1;

  a = current_context->current_stack;
  tmp=m4_transpose(r);
  s4_matconc(a, &r, &tmp);
}

void glRotatef (GLfloat angle, GLfloat x, GLfloat y, GLfloat z)
{
  glRotated (angle, x, y, z);
}

void glTranslated (GLdouble x, GLdouble y, GLdouble z)
{
  Vector3 v;
  Stack4 *s;
  in_begin();
  v = v3_make(x, y, z);
  s = current_context->current_stack;
  s4_translate(s, v);
}

void glTranslatef (GLfloat x, GLfloat y, GLfloat z)
{
  glTranslated (x, y, z);
}

void glScaled (GLdouble x, GLdouble y, GLdouble z)
{
  Vector3 v;
  Stack4 * s;
  in_begin();
  v = v3_make(x, y, z);
  s = current_context -> current_stack;
  s4_scale(s, v);
}

void glScalef (GLfloat x, GLfloat y, GLfloat z)
{
  glScaled (x, y, z);
}


void glMultMatrixd (const GLdouble *m)
{
  Stack4 * s;
  Matrix4 a;
  in_begin();
  a.r1 = v4_make(m[0],m[4],m[8],m[12]);
  a.r2 = v4_make(m[1],m[5],m[9],m[13]);
  a.r3 = v4_make(m[2],m[6],m[10],m[14]);
  a.r4 = v4_make(m[3],m[7],m[11],m[15]);
  s = current_context->current_stack;
  s4_matconc(s, &a, (Matrix4 *)0);

}



/* Replace the current matrix */
void glLoadMatrixd (const GLdouble *m)
{
  Stack4 * s;
  Matrix4 a;
  in_begin();
  a.r1 = v4_make(m[0],m[4],m[8],m[12]);
  a.r2 = v4_make(m[1],m[5],m[9],m[13]);
  a.r3 = v4_make(m[2],m[6],m[10],m[14]);
  a.r4 = v4_make(m[3],m[7],m[11],m[15]);
  s = current_context->current_stack;
  s4_loadmat(s, &a, 0);
}


void glLoadIdentity (void)
{
  Stack4 * s;
  Matrix4  m;
  
  in_begin();
  m = m4_ident();
  s = current_context->current_stack;
  s4_loadmat(s,&m, &m);
}

/* Manipulate the matrix stack */
void glMatrixMode(GLenum mode)
{
  in_begin();
  switch(mode) {
  case GL_MODELVIEW:
    current_context->current_stack = current_context->modelview_stack;
    break;
  case GL_PROJECTION:
    current_context->current_stack = current_context->projection_stack;
    break;
  case GL_TEXTURE:
    current_context->current_stack = current_context->texture_stack;
    break;
  case GL_COLOR:
    current_context->current_stack = current_context->color_stack;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}

void glPushMatrix(void)
{
  Stack4 * c;
  
  in_begin();
  c = current_context->current_stack;
  if((c->mtop - c->mbot) >= (c->size - 1)) {
    SET_ERROR(GL_STACK_OVERFLOW)
  }
  else s4_push(current_context->current_stack);
}


void glPopMatrix(void)
{
  Stack4 * c;
  
  in_begin();
  c = current_context->current_stack;
  if(c->mtop <= c->mbot) {
    SET_ERROR(GL_STACK_UNDERFLOW)
  }
  else s4_pop(current_context->current_stack);
}

/* Specify the viewport */
void glDepthRange (GLclampd near_val, GLclampd far_val)
{
  in_begin();
  current_context->n = near_val;
  current_context->f = far_val;
  
}

void glViewport (GLint x, GLint y, GLsizei width, GLsizei height)
{
  in_begin();
  if(width<0 || height<0) {
    SET_ERROR(GL_INVALID_VALUE)
  }
  else {
    current_context->ox = (GLint)(x + width/2);
    current_context->oy = (GLint)(y + height/2);
    current_context->w  = width;
    current_context->h  = height;
  }
}

void glFrustum (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top,GLdouble near, GLdouble far)
{
  Stack4 * s;
  Matrix4  m;

  in_begin();
  if(near<=0 || far<=0 || left == right || bottom == top || near == far) {
    SET_ERROR(GL_INVALID_VALUE)
  }
  m = m4_ident();
  m.r1.x = 2*near/(right - left);
  m.r1.z = (right + left)/(right - left);
  m.r2.y = 2*near/(top - bottom);
  m.r2.z = - (top + bottom)/(top - bottom);
  m.r3.z = - (far + near)/(far - near);
  m.r3.w = -2*far*near/(far - near);
  m.r4.z = -1;
  m.r4.w = 0;
  s = current_context->current_stack;
  s4_matconc(s, &m, (Matrix4*)0);
}

void glOrtho (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble near, GLdouble far)
{
  Stack4 * s;
  Matrix4 m;

  in_begin();
  if(left == right || bottom == top || near == far){
    SET_ERROR(GL_INVALID_VALUE)
  }
  m = m4_ident();
  m.r1.x = 2/(right - left);
  m.r1.w = -(right + left)/(right - left);
  m.r2.y = 2/(top - bottom);
  m.r2.w = - (top + bottom)/(top - bottom);
  m.r3.z = -2/(far - near);
  m.r3.w = - (far + near)/(far - near);
  s = current_context->current_stack;
  s4_matconc(s, &m, &m);
}



/* Set the current color, color index, or normal vector */
void glColor3d (GLdouble red, GLdouble green, GLdouble blue)
{
  current_context->current_color = v3_make(red, green, blue);
}

void glColor3dv (const GLdouble *v)
{
  glColor3d(v[0],v[1],v[2]);
}

void glColor3f (GLfloat red, GLfloat green, GLfloat  blue)
{
  glColor3d(red, green, blue);
}

void glColor3fv (const GLfloat *v)
{
  glColor3d(v[0],v[1],v[2]);
}

void glNormal3d (GLdouble nx, GLdouble ny, GLdouble nz)
{
  current_context->current_normal = v3_make(nx, ny, nz);
}

void glNormal3dv (const GLdouble *v)
{
  glNormal3d(v[0],v[1],v[2]);
}

void glNormal3f (GLfloat nx, GLfloat ny, GLfloat nz)
{
  glNormal3d(nx, ny, nz);
}

void glNormal3fv (const GLfloat *v)
{
  glNormal3d(v[0],v[1],v[2]);
}

/* Specify light source, material, or lighting model parameter values */
void glLightModelf (GLenum pname, GLfloat param)
{
  in_begin();
  switch(pname){
  case GL_LIGHT_MODEL_LOCAL_VIEWER:
    current_context->local_viewer = (param==0.0)? GL_FALSE: GL_TRUE;
    break;
  case GL_LIGHT_MODEL_TWO_SIDE:
    current_context->two_side =  param;
    break;
  case GL_LIGHT_MODEL_COLOR_CONTROL: 
    current_context->color_control = (GLenum)param; 
    break; 
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}

void glLightModelfv (GLenum pname, const GLfloat *params)
{
  in_begin();
  switch(pname){
  case GL_LIGHT_MODEL_AMBIENT:
    current_context->ambient_scene = v3_make(params[0], params[1], params[2]);
    break;
  case GL_LIGHT_MODEL_LOCAL_VIEWER:
   current_context->local_viewer = (params[0]==0.0)? GL_FALSE: GL_TRUE;
    break;
  case GL_LIGHT_MODEL_TWO_SIDE:
    current_context->two_side = params[0];
    break;
  case GL_LIGHT_MODEL_COLOR_CONTROL: 
    current_context->color_control = (GLenum)params[0]; 
    break;
  default:
    SET_ERROR(GL_INVALID_VALUE)
    break;
  }
}

void glLightf (GLenum light, GLenum pname, GLfloat param)
{
  in_begin();
  if((light - GL_LIGHT0)<0 || (light - GL_LIGHT0)>MAX_LIGHTS) {
    SET_ERROR(GL_INVALID_ENUM)
  }
  else {
    switch(pname){
    case GL_SPOT_EXPONENT:
      if(param<0.0 || param > 180.0){
	SET_ERROR(GL_INVALID_VALUE)
      }
      current_context->l[light - GL_LIGHT0].spot_exponent = param;
      break;
    case GL_SPOT_CUTOFF:
      if( param == 180.0 || (param >0 && param <90.0))
	current_context->l[light - GL_LIGHT0].spot_cutoff = param;
      else {
	SET_ERROR(GL_INVALID_VALUE)
      }
      break;
    case GL_CONSTANT_ATTENUATION:
      if (param<0) {
	SET_ERROR(GL_INVALID_VALUE)
      }
      else current_context->l[light - GL_LIGHT0].const_att = param;
      break;
    case GL_LINEAR_ATTENUATION:
      if (param<0) {
	SET_ERROR(GL_INVALID_VALUE)
      }
      else current_context->l[light - GL_LIGHT0].linear_att = param;
      break;
    case GL_QUADRATIC_ATTENUATION:
      if (param<0) {
	SET_ERROR(GL_INVALID_VALUE)
      }
      else current_context->l[light - GL_LIGHT0].quad_att = param;
      break;
    default:
      SET_ERROR(GL_INVALID_ENUM)
      break;
    }
  }
}

void glLightfv (GLenum light, GLenum pname, const GLfloat *params)
{
  in_begin();
  if((light - GL_LIGHT0)<0 || (light - GL_LIGHT0)>MAX_LIGHTS) {
    SET_ERROR(GL_INVALID_ENUM)
  }
  else {
    switch(pname){
    case GL_AMBIENT:
      current_context->l[light - GL_LIGHT0].ambient = v3_make(params[0], params[1], params[2]);
      break;
    case GL_DIFFUSE:
      current_context->l[light - GL_LIGHT0].diffuse = v3_make(params[0], params[1], params[2]);
      break;
    case GL_SPECULAR:
      current_context->l[light - GL_LIGHT0].specular = v3_make(params[0], params[1], params[2]);
      break;
    case GL_POSITION:
      current_context->l[light - GL_LIGHT0].position = s4_v4xform(current_context->modelview_stack, v4_make(params[0], params[1], params[2], params[3]));
      break;
    case GL_SPOT_DIRECTION:
      current_context->l[light - GL_LIGHT0].spot_direction = s4_n3xform2(current_context->modelview_stack, v3_make(params[0], params[1], params[2]));
      break;
    case GL_SPOT_EXPONENT:
      current_context->l[light - GL_LIGHT0].spot_exponent = params[0];
      break;
    case GL_SPOT_CUTOFF:
      current_context->l[light - GL_LIGHT0].spot_cutoff = params[0];
      break;
    case GL_CONSTANT_ATTENUATION:
      current_context->l[light - GL_LIGHT0].const_att = params[0];
      break;
    case GL_LINEAR_ATTENUATION:
      current_context->l[light - GL_LIGHT0].linear_att = params[0];
      break;
    case GL_QUADRATIC_ATTENUATION:
      current_context->l[light - GL_LIGHT0].quad_att = params[0];
    break;
    default:
      SET_ERROR(GL_INVALID_ENUM)
      break;
    }
  }
}

void glMaterialf (GLenum face, GLenum pname, GLfloat param)
{

  switch(face) {
  case GL_FRONT:
  case GL_BACK:
  case GL_FRONT_AND_BACK:
    current_context->material.face_material = face;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }

  switch(pname){
  case GL_SHININESS:
    if(param<0 || param>128){
      SET_ERROR(GL_INVALID_VALUE)
    }
    else current_context->material.specular_exponent = param;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}

void glMaterialfv (GLenum face, GLenum pname, const GLfloat *params)
{

  switch(face) {
  case GL_FRONT:
  case GL_BACK:
  case GL_FRONT_AND_BACK:
    current_context->material.face_material = face;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
      break;
  }

  switch(pname){
  case GL_AMBIENT:
    current_context->material.ambient = v3_make(params[0], params[1], params[2]);
    break;
  case GL_DIFFUSE:
    current_context->material.diffuse = v3_make(params[0], params[1], params[2]);
    break;
  case GL_AMBIENT_AND_DIFFUSE:
    current_context->material.ambient = v3_make(params[0], params[1], params[2]);
    current_context->material.diffuse = v3_make(params[0], params[1], params[2]);
    break;
  case GL_SPECULAR:
    current_context->material.specular = v3_make(params[0], params[1], params[2]);
    break;
  case GL_EMISSION:
    current_context->material.emissive = v3_make(params[0], params[1], params[2]);
    break;
  case GL_SHININESS:
    if(params[0]<0 || params[0]>128){
      SET_ERROR(GL_INVALID_VALUE)
    }
    else current_context->material.specular_exponent = params[0];
    break;
  case GL_COLOR_INDEXES:
    current_context->material.ambient_index = params[0];
    current_context->material.diffuse_index = params[1];
    current_context->material.specular_index = params[2];
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}



/* Choose a shading model */
void glShadeModel (GLenum mode)
{
  in_begin();
  switch(mode){
  case GL_SMOOTH:
    current_context->smooth_enabled = GL_TRUE;
    break;
  case GL_FLAT:
    current_context->smooth_enabled = GL_FALSE;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}

/* Choose how polygons are rasterized */
void glEdgeFlag( GLboolean flag )
{
  if(flag == 0)
    current_context->edge_flag = GL_FALSE;
  else
    current_context->edge_flag = GL_TRUE;
}

void glEdgeFlagv( const GLboolean *flag )
{
  if(flag[0] == 0)
    current_context->edge_flag = GL_FALSE;
  else
    current_context->edge_flag = GL_TRUE;
}



/* Specify which polygon orientation is a front-facing */
void glFrontFace (GLenum mode) {
  in_begin();
  switch(mode) {
  case GL_CCW:
    current_context->front_sign = 1.0;
    break;
 case GL_CW:
    current_context->front_sign = -1.0;
  break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}


void glCullFace (GLenum mode)
{
  in_begin();
  switch(mode) {
  case GL_FRONT:
    current_context->cull_sign = 1.0;
    break;
  case GL_BACK:
    current_context->cull_sign = -1.0;
    break;
  case GL_FRONT_AND_BACK:
    current_context->cull_sign = 0.0;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
      break;
  }
}


/* Cause a material color to track the current color */
void glColorMaterial (GLenum face, GLenum mode)
{
  in_begin();
  switch(face) {
  case GL_FRONT:
  case GL_BACK:
  case GL_FRONT_AND_BACK:
    current_context->material.face_color_material = face;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
      break;
  }

  switch(mode) {
  case GL_AMBIENT:
  case GL_DIFFUSE:
  case GL_EMISSION:
  case GL_SPECULAR:
  case GL_AMBIENT_AND_DIFFUSE:
    current_context->material.color_material_mode = mode;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}

  /* Obtain light source or material parameter values */
void glGetLightfv (GLenum light, GLenum pname, GLfloat *params)
{
  in_begin();
    
  if((light - GL_LIGHT0)<0 || (light - GL_LIGHT0)>MAX_LIGHTS) {
    SET_ERROR(GL_INVALID_ENUM)
  }

  switch(pname){
  case GL_AMBIENT:
    params[0] = current_context->l[light - GL_LIGHT0].ambient.x;
    params[1] = current_context->l[light - GL_LIGHT0].ambient.y;
    params[2] = current_context->l[light - GL_LIGHT0].ambient.z;
    break;
  case GL_DIFFUSE:
    params[0] = current_context->l[light - GL_LIGHT0].diffuse.x;
    params[1] = current_context->l[light - GL_LIGHT0].diffuse.y;
    params[2] = current_context->l[light - GL_LIGHT0].diffuse.z;
    break;
  case GL_SPECULAR:
    params[0] = current_context->l[light - GL_LIGHT0].specular.x;
    params[1] = current_context->l[light - GL_LIGHT0].specular.y;
    params[2] = current_context->l[light - GL_LIGHT0].specular.z;
    break;
  case GL_POSITION:
    params[0] = current_context->l[light - GL_LIGHT0].position.x;
    params[1] = current_context->l[light - GL_LIGHT0].position.y;
    params[2] = current_context->l[light - GL_LIGHT0].position.z;
    params[3] = current_context->l[light - GL_LIGHT0].position.w;
    break;
  case GL_SPOT_DIRECTION:
    params[0] = current_context->l[light - GL_LIGHT0].spot_direction.x;
    params[1] = current_context->l[light - GL_LIGHT0].spot_direction.y;
    params[2] = current_context->l[light - GL_LIGHT0].spot_direction.z;
    break;
  case GL_SPOT_EXPONENT:
    params[0] = current_context->l[light - GL_LIGHT0].spot_exponent;
    break;
  case GL_SPOT_CUTOFF:
    params[0] = current_context->l[light - GL_LIGHT0].spot_cutoff;
    break;
  case GL_CONSTANT_ATTENUATION:
    params[0] = current_context->l[light - GL_LIGHT0].const_att;
     break;
  case GL_LINEAR_ATTENUATION:
    params[0] = current_context->l[light - GL_LIGHT0].linear_att;
    break;
  case GL_QUADRATIC_ATTENUATION:
    params[0] = current_context->l[light - GL_LIGHT0].quad_att;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}

void glGetMaterialfv(GLenum face, GLenum pname, GLfloat * params)
{
  in_begin();
  /* FALTA "FACE" */
  
  switch(pname){
  case GL_AMBIENT:
    params[0] = current_context->material.ambient.x;
    params[1] = current_context->material.ambient.y;
    params[2] = current_context->material.ambient.z;
    break;
  case GL_DIFFUSE:
    params[0] = current_context->material.diffuse.x;
    params[1] = current_context->material.diffuse.y;
    params[2] = current_context->material.diffuse.z;
    break;
  case GL_SPECULAR:
    params[0] = current_context->material.specular.x;
    params[1] = current_context->material.specular.y;
    params[2] = current_context->material.specular.z;
    break;
  case GL_EMISSION:
    params[0] = current_context->material.emissive.x;
    params[1] = current_context->material.emissive.y;
    params[2] = current_context->material.emissive.z;
    break;
  case GL_SHININESS:
    params[0] = current_context->material.specular_exponent;
    break;
  case GL_COLOR_INDEXES:
    params[0] = current_context->material.ambient_index;
    params[1] = current_context->material.diffuse_index;
    params[2] = current_context->material.specular_index;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}


/* Select the source for pixel reads and copies */
void glReadBuffer (GLenum mode)
{
}

/* Read and write pixels */
void glReadPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels)
{
}

void glDrawPixels (GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels)
{
  in_begin();
}



/* Control per-fragment testing */
void glDepthFunc (GLenum func) {
  in_begin();
}

/* Clear some or all buffers */
void glClear (GLbitfield mask) {
  in_begin();
  if(mask & GL_COLOR_BUFFER_BIT) 
    img_clear(current_context->img,c_scale(256.0, current_context->clear_color));
  if(mask & GL_DEPTH_BUFFER_BIT)
    zbuf_clear(MAX_FLOAT);
}

/* Specify color, depth and stencil values for clears */
void glClearColor (GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha) {
  in_begin();
  current_context->clear_color.x = red;
  current_context->clear_color.y = green;
  current_context->clear_color.z = blue;
}

void glClearDepth (GLclampd depth) {
  in_begin();
  current_context->clear_depth = depth;
}



/* Modes and Execution */
void glEnable (GLenum cap) {
  in_begin();
  switch(cap){
  case GL_NORMALIZE:
    current_context->normalize_enabled = GL_TRUE;
    break;
  case GL_RESCALE_NORMAL:
    current_context->rescale_normal_enabled = GL_TRUE;
    break;
  case GL_SMOOTH:
    current_context->smooth_enabled = GL_TRUE;
    break;
  case GL_FLAT:
    current_context->smooth_enabled = GL_FALSE;
    break;
  case GL_DEPTH_TEST:
    current_context->depth_test_enabled = GL_TRUE;
    break;
  case GL_LIGHTING:
    current_context->lighting_enabled = GL_TRUE;
    break;
  case GL_LIGHT0:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_LIGHT1:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_LIGHT2:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_LIGHT3:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_LIGHT4:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_LIGHT5:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_LIGHT6:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_LIGHT7:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_TRUE;
    break;
  case GL_COLOR_MATERIAL:
    current_context->color_material_enabled = GL_TRUE;
    break;
  case GL_CULL_FACE:
    current_context->culling_enabled = GL_TRUE;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
  }
}

void glDisable (GLenum cap) {
  in_begin();
  switch(cap){
  case GL_NORMALIZE:
    current_context->normalize_enabled = GL_FALSE;
    break;
  case GL_RESCALE_NORMAL:
    current_context->rescale_normal_enabled = GL_FALSE;
    break;
  case GL_SMOOTH:
    current_context->smooth_enabled = GL_FALSE;
    break;
  case GL_FLAT:
    current_context->smooth_enabled = GL_TRUE;
    break;
  case GL_DEPTH_TEST:
    current_context->depth_test_enabled = GL_FALSE;
    break;
  case GL_LIGHTING:
    current_context->lighting_enabled = GL_FALSE;
    break;
  case GL_LIGHT0:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_LIGHT1:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_LIGHT2:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_LIGHT3:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_LIGHT4:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_LIGHT5:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_LIGHT6:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_LIGHT7:
    current_context->light_enabled[cap - GL_LIGHT0] = GL_FALSE;
    break;
  case GL_COLOR_MATERIAL:
    current_context->color_material_enabled = GL_FALSE;
    break;
  case GL_CULL_FACE:
    current_context->culling_enabled = GL_FALSE;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM)
    break;
    
  }

}

GLboolean glIsEnabled (GLenum cap) {
  in_begin();
  switch(cap){
  case GL_NORMALIZE:
    return current_context->normalize_enabled;
    break;
  case GL_RESCALE_NORMAL:
    return current_context->rescale_normal_enabled;
    break;
  case GL_SMOOTH:
    return current_context->smooth_enabled;
    break;
  case GL_FLAT:
    return !current_context->smooth_enabled;
    break;
  case GL_DEPTH_TEST:
    return current_context->depth_test_enabled;
    break;
  case GL_LIGHTING:
    return current_context->lighting_enabled;
    break;
   case GL_LIGHT0:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break; 
  case GL_LIGHT1:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break;
  case GL_LIGHT2:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break;
  case GL_LIGHT3:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break;
  case GL_LIGHT4:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break;
  case GL_LIGHT5:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break;
  case GL_LIGHT6:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break;
  case GL_LIGHT7:
    return current_context->light_enabled[cap - GL_LIGHT0];
    break;
  case GL_COLOR_MATERIAL:
    return current_context->color_material_enabled;
    break;
  case GL_CULL_FACE:
    return current_context->culling_enabled;
    break;
  default:
    SET_ERROR(GL_INVALID_ENUM);
    return 0;
    break;
  }
  
}

/* Wait until all OpenGL comands have executed completely */
void glFinish (void) {
  in_begin();
}

/*  Force all issued OpenGl comands to be executes */
void glFlush (void) {
  in_begin();
}

/* Obtain information about an error or the current OpenGl connection */
GLenum glGetError (void) {
  GLenum e;
  in_begin();
  e = current_context->error;
  current_context->error = GL_NO_ERROR;
  return e;
}

const GLubyte * glGetString (GLenum name) {
  in_begin();
  return "Error";
}




