#ifndef _CONTEXT_H_
#define _CONTEXT_H_

#include <clip.h>
#include <defs.h>
#include <color.h>
#include <geom.h>
#include <laux.h>
#include <list.h>
#include <stack.h> 
#include <image.h>
#include <rasterfile.h>
#include <zbuf.h>

#define MODELVIEW_STACK_SIZE 32
#define PROJECTION_STACK_SIZE 8
#define COLOR_STACK_SIZE 8
#define MAX_LIGHTS 8


typedef struct Vextex_info
{
Vector3 v; /*vertice modelado e projetado */
Vector3 n; /* normal modelada */
Color cpri;   /* cor primaria */
Color csec;   /* cor secundaria */
Vector3 t;   /* textura*/
}Vertex_info;

typedef struct Light
{
Color ambient;          /*acli*/
Color diffuse;          /*dcli*/
Color specular;         /*scli*/
Vector4 position;       /*Ppli*/
Vector3 spot_direction; /*sdli*/
Real spot_exponent;  /*srli*/
Real spot_cutoff;    /*crli*/
Real const_att;
Real linear_att;
Real quad_att;
}Light;


typedef struct Material
{
Color ambient;              /* acm */
Color diffuse;              /* dcm */
Color specular;             /* scm */
Color emissive;             /* ecm */
Real specular_exponent;  /* srm */   /*range: [0.0, 128.0] */
Real ambient_index;      /* am  */
Real diffuse_index;      /* dm  */
Real specular_index;     /* sm  */
unsigned int face_material;
unsigned int face_color_material;
unsigned int color_material_mode;
}Material;


typedef struct GLContext
{
  /**Begin**/
unsigned int begin_mode;
unsigned int into_begin_end;

  /*matrix*/
unsigned int matrix_mode;
Stack4 * modelview_stack;
Stack4 * projection_stack;
Stack4 * texture_stack;
Stack4 * color_stack;
Stack4 * current_stack;

  /**Vertex**/
unsigned char edge_flag;
List * vertex_list;  /* lista de vertex_info */
 
Vector3 current_normal;
Color current_color;
unsigned int color_mode; /* RGBA mode or color index mode */

  
  /**Viewport**/
int w;
int h;
int ox;
int oy;
Real n;
Real f;
  
  /**Enable**/
unsigned char rescale_normal_enabled;
unsigned char normalize_enabled;
unsigned char lighting_enabled;
unsigned char smooth_enabled;
unsigned char flatshading_enabled;
unsigned char depth_test_enabled;
unsigned char color_material_enabled;
unsigned char light_enabled[MAX_LIGHTS];
unsigned char light1_enabled;
unsigned char light2_enabled;
unsigned char light3_enabled;
unsigned char light4_enabled;
unsigned char light5_enabled;
unsigned char light6_enabled;
unsigned char light7_enabled;

  /** Material **/
Material material;

  /** Front Face **/
Real front_sign;

  /** Cull Face **/
unsigned char culling_enabled;
Real cull_sign;

  /** Light Source Parameters **/
Light l[MAX_LIGHTS];

  /** Lighting Model Parameters **/
Color ambient_scene;     /* acs */
unsigned char local_viewer;  /* vbs */
unsigned char two_side;      /* tbs */
unsigned int color_control;    /* ces */

  /**Image**/
Image * img;
Poly * triangle;


/** Clear **/
Vector3 clear_color;
Real clear_index;
Real clear_depth;
int clear_stencil;

/*error*/
unsigned int error;

} GLContext;


GLContext * new_context(int w, int h);
void delete_context(GLContext * c);
GLContext * set_current_context(GLContext * c);
GLContext * get_current_context(void);
void in_begin(void);
void save_context_image(GLContext *c, char *fname);
     
#endif







