#include "illumination.h"



Color material_emission(GLContext * c){
  return c->material.emissive;
}


Color global_ambient_light(GLContext * c){
  return c_mult(c->ambient_scene, c->material.ambient);
}

/* Contribuitions from Light Sources */
Real att_factor(GLContext * c, Item * item, int i)
{
  Real kc, kl, kq;
  if(c->l[i].position.w != 0.0) {
    Real d = v3_dist(VERTEX_INFO(item, v), v3_v4conv(c->l[i].position));
    kc = c->l[i].const_att;
    kl = c->l[i].linear_att;
    kq = c->l[i].quad_att;
    return 1.0/(kc + kl*d + kq*d*d);
  }
  else return 1.0;
}

Real spot_effect(GLContext * c, Item * item, int i)
{
  Vector4 vertex = v4_v3conv(VERTEX_INFO(item, v));
  Real value = v3_max_dot(v4_dir_unit(c->l[i].position, vertex), v3_unit(c->l[i].spot_direction));
  if(c->l[i].spot_cutoff != 180.0) {
    if(value >= cos(DTOR * c->l[i].spot_cutoff))
      return pow(value, c->l[i].spot_exponent);
    else return 0.0;
  }else return 1.0;
}


Color ambient_term(GLContext * c, int i)
{
  return c_mult(c->l[i].ambient, c->material.ambient);
}


Color diffuse_term(GLContext * c, Item * item, int i){
  Vector3 n = VERTEX_INFO(item, n);
  Vector3 L = v4_dir_unit(v4_v3conv(VERTEX_INFO(item, v)), c->l[i].position);
  Real max = v3_max_dot(L, n);
  return c_mult(c_scale(max, c->l[i].diffuse), c->material.diffuse);
}


Color specular_term(GLContext * c, Item * item, int i){
  Vector3 v = VERTEX_INFO(item, v);
  Vector3 n = VERTEX_INFO(item, n);
  Vector3 sum;
  Vector3 L = v4_dir_unit(v4_v3conv(v), c->l[i].position);
  Real spec_exponent = c->material.specular_exponent;
  Color spec_color_light = c->l[i].specular;
  Color spec_material = c->material.specular;

  if(c->local_viewer) {
    Vector4 Pe = v4_make(0,0,0,1);
    sum = v3_unit(v3_add(L, v4_dir_unit(v4_v3conv(v), Pe)));
  }
  else
    sum = v3_unit(v3_add(L, v3_make(0,0,1)));
  
  if(v3_dot(L, n) <= 0) 
    return c_make(0,0,0);  /*nao ha componente especular no vertice */
  else 
    return c_mult(c_scale(pow(v3_max_dot(sum, n),spec_exponent), spec_color_light), spec_material);
}

void single_color(GLContext * c, Item * item)
{
  Color ecm, gal, ambt, dift, spect;
  Color contrib_light = c_make(0, 0, 0);
  Real att;
  Real spot;
  int j;
    
  ecm = material_emission(c);
  gal = global_ambient_light(c);
  
  for(j=0; j<MAX_LIGHTS; j++) {
    if(c->light_enabled[j]) {
      att = att_factor(c, item, j);
      spot = spot_effect(c, item, j);
      ambt = ambient_term(c, j);
      dift = diffuse_term(c, item, j);
      spect = specular_term(c, item, j);
      contrib_light = c_add(contrib_light, c_scale(att * spot , c_add(ambt, c_add(dift, spect))));
    }
  }
  VERTEX_INFO(item, cpri) = c_add(ecm, c_add(gal, contrib_light));
  VERTEX_INFO(item, csec) = c_make (0, 0, 0);
}

void separate_specular_color(GLContext * c, Item * item)
{
  Color ecm, gal, ambt, dift, spect;
  Color contrib_light = c_make(0, 0, 0);
  Color contrib_spec = c_make(0, 0, 0);
  Real att;
  Real spot;
  int j;
    
  ecm = material_emission(c);
  gal = global_ambient_light(c);
  
  for(j=0; j<MAX_LIGHTS; j++) {
    if(c->light_enabled[j]) {
      att = att_factor(c, item, j);
      spot = spot_effect(c, item, j);
      ambt = ambient_term(c, j);
      dift = diffuse_term(c, item, j);
      spect = specular_term(c, item, j);
      contrib_light = c_add(contrib_light,c_scale(att * spot , c_add(ambt, dift)));
      contrib_spec =c_add(contrib_spec, c_scale(att * spot , spect));
    }
  }
  VERTEX_INFO(item, cpri) = c_add(ecm, c_add(gal, contrib_light));;
  VERTEX_INFO(item, csec) = contrib_spec; ;
}







































