#include "gl.h"
#include "rasterization.h"
#include <raster.h>

void transform_normal(GLContext * c, Item * item)
{
  Vector3 n = VERTEX_INFO(item, n);
/*  Vector3 v = VERTEX_INFO(item, v);
  Vector4 u;
  Real q=-v3_dot(n, v);
  u=v4_make(n.x, n.y, n.z, q);
  u=v4_m4mult(u, m4_transpose(*c->modelview_stack->itop));
  n=v3_make(u.x, u.y, u.z);*/
  n = s4_n3xform2(c->modelview_stack, n);
  if (c->rescale_normal_enabled) 
    n = v3_unit(n);
  if (c->normalize_enabled) 
    n = v3_unit(n);
  VERTEX_INFO(item, n) = n;
}


Vector3 viewport_transform(GLContext * c, Vector3 v) {
  Vector3 u;
  u.x = (c->w/2.0)*v.x + c->ox;
  u.y = (c->h/2.0)*v.y + c->oy;
  u.z=((c->f - c->n)/2.0)*v.z+ (c->n + c->f)/2.0;
  return u;
}

void print_vertex_info(GLContext * c) {
  Item * i = LIST_HEAD(c->vertex_list);
  printf("Vertex info\n");
  while(i) {
    Vector3 v;
    v = VERTEX_INFO(i, v);;
    printf("(%f,%f,%f)\n",v.x,v.y,v.z);
    i = NEXT(i);
  }
  printf("\n");
}

void apply_modelview(GLContext * c) {
  Item * i = LIST_HEAD(c->vertex_list);
  while(i) {
    VERTEX_INFO(i, v) = v3_v4conv(s4_v4xform(c->modelview_stack, v4_v3conv(VERTEX_INFO(i, v))));
    i = NEXT(i);
  }
}

void apply_lighting_item(GLContext * c, Item * i) {
  if(c->lighting_enabled) {
    transform_normal(c, i);
    switch(c->color_control) {
    case GL_SINGLE_COLOR:
      single_color(c, i);  
      break;
    case GL_SEPARATE_SPECULAR_COLOR:
      separate_specular_color(c, i);
      break;
    }
  }else return;
}

void apply_lighting_list(GLContext * c) {
  Item * i = LIST_HEAD(c->vertex_list);
  while(i) {
    apply_lighting_item(c, i);
    i = NEXT(i);
  }
}



void apply_projection(GLContext * c) {
  Item * i = LIST_HEAD(c->vertex_list);
  while(i) {
    VERTEX_INFO(i, v) = viewport_transform(c, v3_v4conv(s4_v4xform(c->projection_stack, v4_v3conv(VERTEX_INFO(i, v)))));
    i = NEXT(i);
  }
}

void apply_color(GLContext * c) {
 Item * i = LIST_HEAD(c->vertex_list);
  while(i) {
    VERTEX_INFO(i, cpri) = v3_v4conv(s4_v4xform(c->color_stack, v4_v3conv(VERTEX_INFO(i, cpri))));
    VERTEX_INFO(i, csec) = v3_v4conv(s4_v4xform(c->color_stack, v4_v3conv(VERTEX_INFO(i, csec))));
    i = NEXT(i);
  }  
}

void paint_flatshade(Vector3 v, int n, int lv, Real lt, int rv, Real rt, Real st, void * d) {
  GLContext * c;
  c = (GLContext *)d;
  
  img_putc(c->img, v.x, v.y, c_scale(256.0, COLOR_PRI(c->triangle)->v[2]));
}


void paint_smooth(Vector3 v, int n, int lv, Real lt, int rv, Real rt, Real st, void * d) {
  Vector3 interpolated_color;
  GLContext * c;
  c = (GLContext *)d;
  
  interpolated_color = seg_bilerp(COLOR_PRI(c->triangle), n, st, lv, lt, rv, rt); 
  img_putc(c->img, v.x, v.y, c_scale(256.0,interpolated_color));
}

void paint_flatshade_zbuf(Vector3 v, int n, int lv, Real lt, int rv, Real rt, Real st, void * d) {
  GLContext * c;
  c = (GLContext *)d;
  if(zbuf_store(v))
    img_putc(c->img, v.x, v.y, c_scale(256.0, COLOR_PRI(c->triangle)->v[2]));
}


void paint_smooth_zbuf(Vector3 v, int n, int lv, Real lt, int rv, Real rt, Real st, void * d) {
  Vector3 interpolated_color;
  GLContext * c;
  c = (GLContext *)d;

  if(zbuf_store(v)) {
    interpolated_color = seg_bilerp(COLOR_PRI(c->triangle), n, st, lv, lt, rv, rt); 
    img_putc(c->img, v.x, v.y, c_scale(256.0,interpolated_color));
  }
}

void set_triangle(Poly * p, Vector3 a, Vector3 b, Vector3 c) {
  p->v[0] = a;
  p->v[1] = b;
  p->v[2] = c;
}

int set_index(int j) {
  return !j;
}

Real triangle_area(GLContext * c, Poly * p) {
  Vector3 a, b, d;
  a = p->v[0];
  b = p->v[1];
  d = p->v[2];
  return (c->front_sign)*((a.x*b.y + b.x*d.y + d.x*a.y) - (b.x*a.y + d.x*b.y + a.x*d.y));
}


void raster_triangle(GLContext * c, Item * i0, Item * i1, Item * i2) {

  set_triangle(VERTEX(c->triangle),VERTEX_INFO(i0, v), VERTEX_INFO(i1, v), VERTEX_INFO(i2, v));
  set_triangle(COLOR_PRI(c->triangle), VERTEX_INFO(i0, cpri), VERTEX_INFO(i1, cpri), VERTEX_INFO(i2, cpri));
  set_triangle(COLOR_SEC(c->triangle), VERTEX_INFO(i0, csec), VERTEX_INFO(i1, csec), VERTEX_INFO(i2, csec));

  if((!c->culling_enabled) || (triangle_area(c, VERTEX(c->triangle)) * c->cull_sign < 0))  {
    if(!c->smooth_enabled) {
      if(c->depth_test_enabled)
	scan_poly(c->triangle, paint_flatshade_zbuf, c);
      else
	scan_poly(c->triangle, paint_flatshade, c);
    }
    else {
      if(c->depth_test_enabled)
	scan_poly(c->triangle, paint_smooth_zbuf, c);
      else
	scan_poly(c->triangle, paint_smooth, c);
    }
  }else return;
}


void raster_points(void) {
  GLContext * c = get_current_context();
  Item * item;
  Vector3 vertex;
  Color color;
  
  apply_modelview(c);
  apply_lighting_list(c);
  apply_projection(c);
  apply_color(c);
  item = LIST_HEAD(c->vertex_list);
  while(item) {
    color = VERTEX_INFO(item, cpri);
    vertex = VERTEX_INFO(item, v);
    img_putc(c->img, vertex.x, vertex.y, c_scale(256.0,color));
    item = NEXT(item);
  }
}

void raster_line_strip() {
}
void raster_line_loop() {
}
void raster_lines() {
}

void raster_triangles() {   
  GLContext * c = get_current_context();
  Item * item;

  apply_modelview(c);
  apply_lighting_list(c);
  apply_projection(c);
  apply_color(c);
  item = LIST_HEAD(c->vertex_list);
  
  while(item) {
    if(SECOND(item) && THIRD(item)) {
      raster_triangle(c, item, SECOND(item), THIRD(item));
      item = FOURTH(item);
    }else return;
  }
}

void raster_triangle_strip() {
  GLContext * c = get_current_context();
  Item * current_item;
  Item * item[2];
  int j;

  apply_modelview(c);
  apply_lighting_list(c);
  apply_projection(c);
  apply_color(c);

  current_item = LIST_HEAD(c->vertex_list);
    
  j = 0;
  item[j] = current_item;
 
  if(NEXT(current_item)) {
    j = set_index(j);
    item[j] = NEXT(current_item);
  } else return;
  
  if (NEXT(NEXT(current_item)))
    current_item = NEXT(NEXT(current_item));
  else return;
  
  while(current_item) {
    raster_triangle(c, item[0], item[1], current_item);
    j = set_index(j);
    item[j] = current_item;
    current_item = NEXT(current_item);
  }
}

void raster_triangle_fan() {
  GLContext * c = get_current_context();
  Item * current_item;
  Item * item[2];
  int j;

  apply_modelview(c);
  apply_lighting_list(c);
  apply_projection(c);
  apply_color(c);

  current_item = LIST_HEAD(c->vertex_list);
  
  j = 0;
  item[0] = current_item;
  j = set_index(j);
  if(NEXT(current_item)) {
    item[j] = NEXT(current_item);
  } else return;
  if (NEXT(NEXT(current_item)))
    current_item = NEXT(NEXT(current_item));

  while(current_item) {
    raster_triangle(c, item[0], item[1], current_item);
    item[j] = current_item;
    current_item = NEXT(current_item);
  }
}

void raster_polygon() {
 raster_triangle_fan();
}

void raster_quad_strip() {
  GLContext * c = get_current_context();
  Item * current_item;
  Item *item[2];
  int j;

  apply_modelview(c);
  apply_lighting_list(c);
  apply_projection(c);
  apply_color(c);

  current_item = LIST_HEAD(c->vertex_list);
  
  j = 0;
  item[j] = current_item;
  j = set_index(j);
  item[j] = NEXT(current_item);
  current_item = THIRD(current_item);

  while(current_item) {
    if(NEXT(current_item)) {
      if(!c->smooth_enabled) {
	VERTEX_INFO(item[0],cpri) = VERTEX_INFO(item[1],cpri)
	  = VERTEX_INFO(current_item,cpri) = VERTEX_INFO(NEXT(current_item),cpri); 
      } 
      raster_triangle(c, item[0], item[1], NEXT(current_item));
      raster_triangle(c, NEXT(current_item), current_item, item[0]);
      j = set_index(j);
      item[j] = current_item;
      j = set_index(j);
      item[j] = NEXT(current_item);
      current_item = THIRD(current_item);
      
    }else return;
  }

}

void raster_quads() {
  GLContext * c = get_current_context();
  Item * item;
  
  apply_modelview(c);
  apply_lighting_list(c);
  apply_projection(c);
  apply_color(c);
  item = LIST_HEAD(c->vertex_list);
    
  while (item) {
    if(SECOND(item) && THIRD(item) && FOURTH(item)) {
      if(!c->smooth_enabled) {
	VERTEX_INFO(item,cpri) = VERTEX_INFO(SECOND(item),cpri)
	  = VERTEX_INFO(THIRD(item),cpri) = VERTEX_INFO(FOURTH(item),cpri); 
      } 
      raster_triangle(c, item ,SECOND(item), THIRD(item));
      raster_triangle(c, THIRD(item), FOURTH(item), item);
      item = NEXT(FOURTH(item));
    } else return;
  }
}










