#include <stdio.h>
#include <stdlib.h>
#include <GL/context.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

FILE * teapot_file;
int max_vertices;
GLfloat * vertices;
int max_triangles;
GLint * triangles;

void open_teapot() {
  teapot_file=fopen("teapot.nff", "r");
}

void read_vertices() {
  int i;
  fscanf(teapot_file, "%d", &max_vertices);
  vertices=(GLfloat *)malloc(max_vertices*3*sizeof(GLfloat));
  for(i=0; i<max_vertices; ++i) {
    fscanf(teapot_file, "%f %f %f", &vertices[i*3], &vertices[i*3+1], &vertices[i*3+2]);
  }
}
  
void read_triangles() {
  int i, tmp;
  fscanf(teapot_file, "%d", &max_triangles);
  triangles=(GLint *)malloc(max_triangles*3*sizeof(GLint));
  for(i=0; i<max_triangles; ++i) {
    fscanf(teapot_file, "%d %d %d %d", &tmp, &triangles[i*3], &triangles[i*3+1], &triangles[i*3+2]);
  }
}

void draw(void) 
{
  int i;
  GLfloat mat_ambient[] = { 0.5, 0.5, 0.5, 1.0 };
  GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat mat_shininess[] = { 10.0 };
  GLfloat light_position[] = { 5.0, 5.0, 5.0, 1.0 };
  GLfloat model_ambient[] = { 0.7, 0.2, 0.7, 1.0 };
  GLfloat white_light[] = {1.0, 1.0, 1.0, 1.0 };
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glShadeModel(GL_FLAT);
  
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light );
  //   glLightfv(GL_LIGHT0, GL_SPECULAR, white_light);
  
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, model_ambient);
  
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_NORMALIZE);

  glTranslatef(0.0, -1.6, 0.0);
  glRotatef(-90.0, 1.0, 0.0, 0.0);
  for(i=0; i<max_triangles; ++i) {
    int j;
    Vector3 v[3], n;
    j=triangles[3*i];
    v[0]=v3_make(vertices[3*j], vertices[3*j+1], vertices[3*j+2]);
    j=triangles[3*i+1];
    v[1]=v3_make(vertices[3*j], vertices[3*j+1], vertices[3*j+2]);
    j=triangles[3*i+2];
    v[2]=v3_make(vertices[3*j], vertices[3*j+1], vertices[3*j+2]);
    n=v3_cross(v3_sub(v[1], v[0]), v3_sub(v[2], v[0]));
    glNormal3f(n.x, n.y, n.z);
    glBegin(GL_TRIANGLES);
    glVertex3f(v[0].x, v[0].y, v[0].z);
    glVertex3f(v[1].x, v[1].y, v[1].z);
    glVertex3f(v[2].x, v[2].y, v[2].z);
    glEnd();
  }

}


void reshape (int w, int h)
{
  glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if (w <= h)
    glOrtho(-4.0, 4.0, -4.0*(GLfloat)h/(GLfloat)w,
	    4.0*(GLfloat)h/(GLfloat)w, -10.0, 10.0);
  else
    glOrtho(-4.0*(GLfloat)w/(GLfloat)h,
	    4.0*(GLfloat)w/(GLfloat)h, -4.0, 4.0, -10.0, 10.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}


int main(){
  
  GLContext * c;
  c = new_context(500,500);
  set_current_context(c);
  
  reshape(500, 500);
  open_teapot();
  read_vertices();
  read_triangles();
  draw();

  save_context_image(c,"teapot.ras");
  delete_context(c);
  return 0;
}


















